/*
 * ypp-commodities main program: argument parsing etc.
 */
/*
 *  This is part of ypp-sc-tools, a set of third-party tools for assisting
 *  players of Yohoho Puzzle Pirates.
 * 
 *  Copyright (C) 2009 Ian Jackson <ijackson@chiark.greenend.org.uk>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *  Yohoho and Puzzle Pirates are probably trademarks of Three Rings and
 *  are used without permission.  This program is not endorsed or
 *  sponsored by Three Rings.
 */

#include "convert.h"

void debug_flush(void) {
  sysassert(!ferror(debug));
  sysassert(!fflush(debug));
}

const char *get_vardir(void) { return "."; }

static enum {
  mf_findwindow=      0001,
  mf_screenshot=      0010,
  mf_readscreenshot=  0020,
  mf_analyse=         0100,
  
  mode_findwindow=    0001,
  mode_screenshot=    0011,
  mode_analyse=       0120,

  mode_all=           0111,
} o_mode= mode_all;

static char *o_screenshot_fn;
static int o_single_page, o_quiet;
static const char *o_outputmode= "upload";

const char *o_resolver;
FILE *screenshot_file;


static void vbadusage(const char *fmt, va_list) FMT(1,0) NORET;
static void vbadusage(const char *fmt, va_list al) {
  fputs("bad usage: ",stderr);
  vfprintf(stderr,fmt,al);
  fputc('\n',stderr);
  exit(12);
}
DEFINE_VWRAPPERF(static, badusage, NORET);

static void open_screenshot_file(const char *mode) {
  screenshot_file= fopen(o_screenshot_fn, mode);
  if (!screenshot_file)
    fatal("could not open screenshots file `%s': %s",
	  o_screenshot_fn, strerror(errno));
}

static void run_analysis(void) {
  FILE *tf;

  sysassert( tf= tmpfile() );
  progress("running recognition...");
  analyse(tf);

  if (o_single_page && !strcmp(o_outputmode,"upload"))
    fatal("Recognition successful, but refusing to upload partial data\n"
	  " (--single-page specified).  Specify an output mode?");

  sysassert( fseek(tf,0,SEEK_SET) == 0);

  progress_log("processing results (--%s)...", o_outputmode);
  pid_t processor;
  sysassert( (processor= fork()) != -1 );

  if (!processor) {
    sysassert( dup2(fileno(tf),0) ==0 );
    execlp("./yppsc-commod-processor", "yppsc-commod-processor",
	   o_outputmode, (char*)0);
    sysassert(!"execlp commod-processor failed");
  }

  waitpid_check_exitstatus(processor, "output processor/uploader");
  fclose(tf);
  progress_log("all complete.");
}  

int main(int argc, char **argv) {
  const char *arg;
  int r;

#define ARGVAL  ((*++argv) ? *argv : \
		 (badusage("missing value for option %s",arg),(char*)0))

  while ((arg=*++argv)) {
    if (!strcmp(arg,"--find-window-only"))
      o_mode= mode_findwindow;
    else if (!strcmp(arg,"--screenshot-only"))
      o_mode= mode_screenshot;
    else if (!strcmp(arg,"--analyse-only") ||
	     !strcmp(arg,"--same"))
      o_mode= mode_analyse;
    else if (!strcmp(arg,"--everything"))
      o_mode= mode_all;
    else if (!strcmp(arg,"--single-page"))
      o_single_page= 1;
    else if (!strcmp(arg,"--quiet"))
      o_quiet= 1;
    else if (!strcmp(arg,"--edit-charset"))
      o_resolver= "./yppsc-ocr-resolver";
    else if (!strcmp(arg,"--raw-tsv"))
      o_outputmode= 0;
    else if (!strcmp(arg,"--upload") ||
	     !strcmp(arg,"--arbitrage") ||
	     !strcmp(arg,"--tsv") ||
	     !strcmp(arg,"--best-prices"))
      o_outputmode= arg+2;
    else if (!strcmp(arg,"--screenshot-file"))
      o_screenshot_fn= ARGVAL;
#define DF(f)					\
    else if (!strcmp(arg,"-D" #f))		\
      debug_flags |= dbg_##f;
    DEBUG_FLAG_LIST
#undef DF
    else if (!strcmp(arg,"--window-id")) {
      char *ep;
      unsigned long windowid= strtoul(ARGVAL,&ep,0);
      if (*ep) badusage("invalid window id");
      set_yppclient_window(windowid);
    } else
      badusage("unknown option `%s'",arg);
  }
  
  if (!o_screenshot_fn) {
    r= asprintf(&o_screenshot_fn,"%s/#pages#.ppm",get_vardir());
    sysassert(r>=0);
  }

  if (o_mode & mf_findwindow) {
    screenshot_startup();
    find_yppclient_window();
  }
  if (o_mode & mf_screenshot) {
    open_screenshot_file("w");
    if (o_single_page) take_one_screenshot();
    else take_screenshots();
  }
  if (o_mode & mf_readscreenshot) {
    open_screenshot_file("r");
    if (o_single_page) read_one_screenshot();
    else read_screenshots();
  }
  if (o_mode & mf_analyse) {
    if (o_outputmode)
      run_analysis();
    else
      analyse(stdout);
  }
  return 0;
}




DEFINE_VWRAPPERF(, progress, )
DEFINE_VWRAPPERF(, progress_log, )
DEFINE_VWRAPPERF(, progress_spinner, )
DEFINE_VWRAPPERF(, warning, )
DEFINE_VWRAPPERF(, fatal, NORET)

static int last_progress_len;
     
static void vprogress_core(int spinner, const char *fmt, va_list al) {
  int r;
  
  if (o_quiet) return;
  if (!isatty(2)) return;
  
  if (last_progress_len)
    putc('\r',stderr);

  r= vfprintf(stderr,fmt,al);

  if (spinner) {
    putc(spinner,stderr);
    r++;
  }

  if (r < last_progress_len) {
    fprintf(stderr,"%*s", last_progress_len - r, "");
    if (!r) putc('\r', stderr);
    else while (last_progress_len-- > r) putc('\b',stderr);
  }
  last_progress_len= r;

  if (ferror(stderr) || fflush(stderr)) _exit(16);
}
   
void vprogress(const char *fmt, va_list al) { vprogress_core(0,fmt,al); }
void vprogress_spinner(const char *fmt, va_list al) {
  static const char spinchars[]="/-\\";
  static int spinner;

  vprogress_core(spinchars[spinner],fmt,al);
  spinner++;
  spinner %= (sizeof(spinchars)-1);
}

void vprogress_log(const char *fmt, va_list al) {
  if (o_quiet) return;
  
  progress("");
  vfprintf(stderr,fmt,al);
  putc('\n',stderr);
  fflush(stderr);
}

void vwarning(const char *fmt, va_list al) {
  progress("");
  fputs("Warning: ",stderr);
  vfprintf(stderr,fmt,al);
  fputs("\n",stderr);
  fflush(stderr);
}

void vfatal(const char *fmt, va_list al) {
  progress("");
  fputs("\n\nFatal error: ",stderr);
  vfprintf(stderr,fmt,al);
  fflush(stderr);
  fputs("\n\n",stderr);
  _exit(4);
}

void sysassert_fail(const char *file, int line, const char *what) {
  int e= errno;
  progress("");
  fprintf(stderr,
	  "\nfatal operational error:\n"
	  " unsuccessful execution of: %s\n"
	  " %s:%d: %s\n\n",
	  what, file,line, strerror(e));
  _exit(16);
}

void *mmalloc(size_t sz) {
  void *r;
  if (!sz) return 0;
  sysassert( r= malloc(sz) );
  return r;
}
void *mrealloc(void *p, size_t sz) {
  assert(sz);
  void *r;
  sysassert( r= realloc(p,sz) );
  return r;
}

void waitpid_check_exitstatus(pid_t pid, const char *what) { 
  pid_t got;
  int st;
  for (;;) {
    got= waitpid(pid, &st, 0);
    if (pid==-1) { sysassert(errno==EINTR); continue; }
    break;
  }
  sysassert( got==pid );

  if (WIFEXITED(st)) {
    if (WEXITSTATUS(st))
      fatal("%s failed with nonzero exit status %d",
	    what, WEXITSTATUS(st));
  } else if (WIFSIGNALED(st)) {
    fatal("%s died due to signal %s%s", what,
	  strsignal(WTERMSIG(st)), WCOREDUMP(st)?" (core dumped)":"");
  } else {
    fatal("%s gave strange wait status %d", what, st);
  }
}
